"""
Small module to get information from the server config.

:author: Daniel Abercrombie <dabercro@mit.edu>
"""


import os
import sys
import shutil

import yaml


class NoConfig(Exception):
    """
    An exception that is raised if there is no config file
    """

LOCATION = None

def config_dict():
    """
    :returns: the configuration in a dict
    :rtype: str
    :raises NoConfig: when it cannot find the configuration file
    """

    global LOCATION # pylint: disable=global-statement

    if LOCATION is None:
        if os.path.basename(sys.argv[0]) == 'workflowtool' and len(sys.argv) > 1:
            LOCATION = sys.argv[1]
        else:
            LOCATION = 'config.yml'

    if not os.path.exists(LOCATION):

        default_loc = os.path.join(os.path.dirname(__file__),
                                   'default', 'config.yml')


        if os.path.basename(sys.argv[0]) == 'workflowtool':
            shutil.copy(default_loc, LOCATION)
            sys.tracebacklimit = 0
            raise NoConfig(
                '\n\n  Copied a default configuration to %s.\n  '
                'Please check it, and then run "workflowtool" again.\n'
                % LOCATION)

        LOCATION = os.path.join(default_loc)

    output = {}

    with open(LOCATION, 'r') as config:
        output = yaml.load(config, Loader=yaml.FullLoader)

    return output


def get_valid_emails():
    """Get iterator for valid email patterns for this instance.
    This is configurable by the webmaster.

    :returns: List of valid email patterns
    :rtype: list
    """

    valid = config_dict()['valid_emails']

    emails = []
    for domain in valid.get('domains', []):
        emails.append('@' + domain)

    for email in valid.get('whitelist', []):
        emails.append(email)

    return emails


def wm_email():
    """
    :returns: the email of the webmaster
    :rtype: str
    """

    return config_dict()['webmaster']['email']


def wm_name():
    """
    :returns: the name of the webmaster
    :rtype: str
    """

    return config_dict()['webmaster']['name']


#def host_name():
#    """
#    :returns: the name of the host from the config
#    :rtype: str
#    """
#
#    return config_dict()['host']['name']
#
#
#def host_port():
#    """
#    :returns: the port to host the site on
#    :rtype: str
#    """
#
#    return config_dict()['host']['port']
#
#
def workflow_history_path():
    """
    :returns: the path to the files of workflow error history
    :rtype: str
    """

    return config_dict()['data']['workflow_history']

def all_errors_path():
    """
    :returns: the path to all errors for recent workflows
    :rtype: str
    """

    return config_dict()['data']['all_errors']

#def get_cluster_settings():
#    """
#    :returns: dictionary containing the settings for clustering
#    :rtype: dict:
#    """
#
#    return config_dict()['cluster']
#
def get_history_length():
    """
    :returns: the number of days of history to check for workflows
              with an action submitted.
    :rtype: int
    """

    return int(config_dict()['actions']['submithistory'])



import glob
import setuptools

import workflowwebtools

setuptools.setup(
    name='workflowwebtools',
    version=workflowwebtools.__version__,
    packages=setuptools.find_packages(),
    author='Daniel Abercrombie',
    author_email='dabercro@mit.edu',
    description='Provides a server for Production and Reprocessing operations',
    url='https://github.com/CMSCompOps/WorkflowWebTools',
    scripts=[s for s in glob.glob('bin/*') if not s.endswith('~')],
    package_data={
        'workflowwebtools': ['default/config.yml'],
        'workflowwebtools.web': ['static/*/*',
                                 'templates/*']
        },
    install_requires=[
        'requests',
        'cmstoolbox>=0.15.1',
        'CMSMonitoring',
        'more-itertools',
        'cherrypy',
        'mako',
        'numpy',
        'scikit-learn',
        'passlib',
        'bcrypt',
        'pyOpenSSL',
        'pytest>=4.4.2'
        'pyyaml>=5.1',
        'validators',
        'tabulate',
        'pymongo<3.5.0',
        'cx_Oracle'
        ]
    )



"""
The :py:mod:`actionshistorylink` module creates files that link
the action and the errors thrown for each task where the action and history
are both stored on the server.
The ``config.yml`` file is read to determine the locations of this history databases.

:author: Daniel Abercrombie <dabercro@mit.edu>
"""

import json

from . import serverconfig
from . import manageactions
from . import globalerrors


def dump_json(file_name=None):
    """
    Dump a list of pairs into a file and returns the dictionary.
    The pairs are dictionary of errors, and action document.
    Each element in the list corresponds to a different subtask.

    :param str file_name: The location to place the json file, if set
    :returns: The errors and actions for each subtask
    :rtype: dict
    """

    output = {}

    history = globalerrors.ErrorInfo(serverconfig.workflow_history_path())
    actions = manageactions.get_actions(0, acted=None)

    session = {'info': history}

    for workflow in actions:
        for subtask in history.get_step_list(workflow):
            action = actions[workflow]['Action']
            if action in ['acdc', 'recovery']:
                parameters = actions[workflow]['Parameters'].get(
                    '/'.join(subtask.split('/')[2:]), {})
            else:
                parameters = actions[workflow]['Parameters']

            output[subtask] = {
                'errors': {
                    'good_sites': globalerrors.get_step_table(
                        subtask, session, readymatch=['green'],
                        sparse=True),
                    'bad_sites': globalerrors.get_step_table(
                        subtask, session, readymatch=['yellow', 'red', 'none'],
                        sparse=True)
                    },
                'parameters':
                    parameters
                }

            output[subtask]['parameters']['action'] = \
                action if action != 'special' else \
                actions[workflow]['Parameters']['action']

    if file_name:
        with open(file_name, 'w') as output_file:
            json.dump(output, output_file)

    return output



"""
This module reads the explain error info for a session and tries to
classify them in a way that can help recommend procedures to the operator.
These procedures are gathered from :py:mod:`WorkflowWebTools.procedures`.

:author: Daniel Abercrombie <dabercro@mit.edu>
"""

import re

from collections import defaultdict

from .procedures import PROCEDURES

def classifyerror(errorcode, workflow):
    """
    Return the most relevant characteristics of an error code for this session.
    This will include things like:

    - Files not opening
    - StageOut errors

    .. Note::

       More error types should be added to this function as needed

    :param int errorcode: The error code that we want to classify
    :param workflowinfo.WorkflowInfo workflow: the workflow that we want to get the errors from
    :returns: A tuple of strings describing the key characteristics of the errorcode.
              These strings are good for printing directly in web browsers.
              The first string is the types of errors reported with this error code.
              The second string is the recommended normal actions.
              The third string contains special parameters useful for additional actions.
    :rtype: str, str, str
    """

    procedure = PROCEDURES.get(errorcode, {})

    logs = workflow.get_explanation(str(errorcode))

    error_re = re.compile(r'[\w\s]+ \(Exit code: (\d+)\)')
    error_types = {}

    additional_re = procedure.get('additional', {}).get('re', None)
    additional_params = []

    for log in logs:
        for line in log.split('\n'):
            # Add each type of error associated with the log
            match = error_re.match(line)
            if match and match.group(1) not in error_types.keys():
                error_types[match.group(1)] = match.group(0)

            # Get additional parameters
            if additional_re:
                add_match = additional_re.search(line)
                if add_match and add_match.group(1) not in additional_params:
                    additional_params.append(add_match.group(1))

    # Create the strings to return

    error_types_string = '<br>'.join([error_types[key] for key in \
                                          sorted(error_types.keys(), key=int)])

    normal_action_string = procedure.get('normal', 'No instructions for this error code')

    additional_actions_string = '<br>'.join(
        [PROCEDURES[errorcode].get('additional', {}).get('action', ''), 'Problems:'] +
        additional_params) if additional_params else ''

    # The procedures use ' |br| |br| ' to break lines because sphinx uses that
    # to replace with raw html

    return {'types': error_types_string,
            'recommended': normal_action_string.replace(' |br| |br| ', '<br>'),
            'params_string': additional_actions_string.replace(' |br| |br| ', '<br>'),
            'list_of_params': additional_params}


def get_max_errorcode(workflow):
    """
    Get the errorcode with the most errors for a session

    :param workflowinfo.WorkflowInfo workflow: the workflow that we want to get the errors from
    :returns: The error code that appears most often for this workflow
    :rtype: int
    """

    errors = workflow.get_errors(True)
    errors_summed = defaultdict(int)

    for codes in errors.values():
        for errorcode, sites in codes.items():
            numcode = -1 if errorcode == 'NotReported' else int(errorcode)
            for num in sites.values():
                errors_summed[numcode] += num

    output = 0
    max_num = 0
    for code, num in errors_summed.items():
        if num > max_num:
            max_num = num
            output = code

    return output



# pylint: disable=redefined-builtin

"""
Here is the list of recommended procedures for various exit codes:

%s

The module :py:mod:`WorkflowWebTools.procedures`, which is used by
:py:mod:`WorkflowWebTools.classifyerrors` contains only the ``PROCEDURES`` dictionary.
The keys of this dictionary are the integer exit codes from the agent to act on.
The values are dictionaries that contain ``'normal'`` and
``'additional'`` actions and a ``'cause'``.
The ``'additional'`` actions are dictionaries containing ``'re'`` and ``'action'``.
The ``'re'`` field is a compiled regular expression to match to the logs.
``group(1)`` of this regular expression is taken to be a parameter to be displayed to the operator.
The ``'action'`` field gives additional instructions to the operator,
who should inspect the parameters returned by the regular expression to determine
the ultimate action to perform.
The ``'cause'`` will add some insight into the error codes in the documentation.
It isn't shown to the operator because they already see the different types of error codes
and are provided with a link to look at the error logs directly.

An example procedure with just exit code ``84`` would look like the following::

    PROCEDURES = {
        84: {
            'normal': 'ACDC',
            'cause': 'The file could not be found',
            'additional': {
                're': re.compile(r'(root://.+root)'),
                'action': ('If a second round of ACDC and fails for same files, '
                           'contact the transfer team and/or site support team.')
                }
            }
        }

What this means is that error code 84 usually should be an ACDC.
However, if the operator sees the same files appear from the regular expression,
then the operator should contact the transfer team and/or the site support team.

These dictionaries are used to generate the procedures table above.

:author: Daniel Abercrombie <dabercro@mit.edu>
"""

import re
import textwrap

from tabulate import tabulate # pylint: disable=import-error

#
# These are the procedures listed on the WTC twiki:
# https://twiki.cern.ch/twiki/bin/view/CMS/CompOpsPRWorkflowTrafficController#Error_Codes_Mapping
# The normal procedure for each error code is listed
# Additional procedures take into account some parameters
# The parameters are extracted into group(1) of the regular expressions
#
PROCEDURES = {
    84: {
        'normal': 'ACDC',
        'cause': 'The file could not be found',
        'additional': {
            're': re.compile(r'(root://[-/\w\.]+\.root)'),
            'action': ('If a second round of ACDC and fails for same files, '
                       'contact the transfer team and/or site support team.')
            }
        },
    85: {
        'normal': 'ACDC',
        'cause': ('If FileReadError, the file could not be read. |br| |br| '
                  'If R__unzip: error, most likely a bad node with corrupted '
                  'intermediate files.'),
        'additional': {
            're': re.compile(r'(root://[-/\w\.]+\.root|R__unzip)'),
            'action': ('If getting unzip errors, most likely a bad node. '
                       'Report immediately. |br| |br| '
                       'If a second round of ACDC and fails for same files, '
                       'contact the transfer team and/or site support team.')
            }
        },
    92: {
        'normal': 'ACDC',
        'cause': 'File was not found and fallback procedure also failed.',
        'additional': {
            're': re.compile(r'(root://[-/\w\.]+\.root)'),
            'action': ('If a second round of ACDC and fails for same files, '
                       'contact the transfer team and/or site support team.')
            }
        },
    134: {
        'normal': 'Strong failure to report immediately'
        },
    137: {
        'normal': 'Point SSP. Open GGUS ticket and put site in drain.',
        'cause': 'Likely an unrelated batch system kill. Sometimes a site-related problem.'
        },
    99109: {
        'normal': ('Check the Site Readiness and consult with the Site Support Team. |br| |br| '
                   'If a temporary failure: ACDC. |br| |br| '
                   'If a problem with the site: Make GGUS ticket, put the site in drain '
                   'then kill and clone'),
        'cause': 'Uncaught exception in WMAgent step executor (often staging out problems)'
        },
    8004: {
        'normal': 'Report immediately to the Site Support Team.',
        'cause': 'BadAlloc - memory exhaustion'
        },
    11003: {
        'normal': 'ACDC',
        'cause': 'JobExtraction failures'
        },
    73: {
        'normal': 'Open a GitHub issue',
        'cause': ('RootEmbeddedFileSequence: secondary file list seems to be missing. |br| |br| '
                  'When overflow procedure is enables, the agent should be able to see the '
                  'secondary file set as present at those sites as well.')
        },
    50660: {
        'normal': 'More memory',
        'cause': 'Performance kill: "Job exceeding maxRSS"'
        },
    71304: {
        'normal': 'ACDC',
        'cause': 'Job was killed by WMAgent for using too much wallclock time'
        },
    71305: {
        'normal': 'ACDC',
        'cause': 'Job was killed by WMAgent for using too much wallclock time'
        },
    71306: {
        'normal': 'ACDC',
        'cause': 'Job was killed by WMAgent for using too much wallclock time'
        },
    61202: {
        'normal': 'If temporary failure: ACDC, otherwise contact glideinWMS team',
        'cause': 'Can\'t determine proxy filename. X509 user proxy required for job.'
        },
    50513: {
        'normal': 'ACDC',
        'cause': 'Failure to run SCRAM setup scripts'
        },
    99303: {
        'normal': 'Strong failure, report immediately',
        'cause': ('The agent has not found the JobReport.x.pkl file. '
                  'Regardless of the status of the job, it will fail the job with 99303. '
                  'Work is needed to make this error as rare as possible since there '
                  'are very limited cases to make the file fail to appear.')
        },
    8020: {
        'normal': 'ACDC',
        'cause': 'File was not found and fallback procedure also failed.',
        'additional': {
            're': re.compile(r'(root://[-/\w\.]+\.root)'),
            'action': ('Check that the file exists')
            }
        },
    8021: {
        'normal': 'ACDC',
        'cause': 'File was not found and fallback procedure also failed.',
        'additional': {
            're': re.compile(r'(root://[-/\w\.]+\.root)'),
            'action': ('Check that the file exists')
            }
        },
    8028: {
        'normal': 'ACDC',
        'cause': 'File was not found and fallback procedure also failed.',
        'additional': {
            're': re.compile(r'(root://[-/\w\.]+\.root)'),
            'action': ('Check that the file exists')
            }
        },
    8001: {
        'normal': 'Strong failure, report immediately'
        },
    8002: {
        'normal': 'Strong failure, report immediately'
        },
    71104: {
        'normal': ('Copy input files from tape (?) or '
                   'wait for site listed to come out of drain.'),
        'cause': 'Site containing only available copy of input datasets is in drain.',
        'additional': {
            're': re.compile(r'The job can run only at .*(T[12].*)')
            }
        }
    }

# Here is a list comprehension to make you cry.
# I'm basically trying to insert line breaks into the long fields,
# while reseting textwrap's counts at manual line breaks

WRAP = lambda text: ' |br| '.join(
    [' |br| '.join(textwrap.wrap(line, 25)) \
         for line in text.split(' |br| ')]
    )

__doc__ %= tabulate(
    [
        [key,
         WRAP(item.get('cause', '')),
         WRAP(item['normal']),
         WRAP(item.get('additional', {}).get('action', '')),
         ' |br| '.join(item['additional']['re'].pattern.strip('()').split('|')) \
             if item.get('additional', {}).get('re') else '',
        ] \
            for key, item in sorted(PROCEDURES.items(), key=lambda x: x[0])],
    headers=['Exit |br| Code', 'Cause', 'Normal Procedure',
             'Additional Procedure', 'Additional Match |br| Expression'],
    tablefmt='grid'
)



"""Module for manipulating the database that stores past operator reasons.

:author: Daniel Abercrombie <dabercro@mit.edu>
"""

import os
import sqlite3
import cherrypy

from . import serverconfig

DEFAULT_SHORT = '---- No Short Reason Given, Not Saved to Database! ----'


def get_reasons():
    """Gets the reasons database in the local directory.

    :returns: the reasons connection, cursor
    :rtype: (sqlite3.Connection, sqlite3.Cursor)
    """

    conn = sqlite3.connect(os.path.join(serverconfig.config_dict()['workspace'], 'reasons.db'))
    curs = conn.cursor()
    curs.execute('SELECT name FROM sqlite_master WHERE type="table" and name="reasons"')

    if not curs.fetchone():
        curs.execute('CREATE TABLE reasons (shortreason varchar(255) PRIMARY KEY, '
                     'longreason varchar(4095))')

    return conn, curs


def update_reasons(reasons):
    """Gets the reasons for a given action and updates the reasons db

    :param reasons: is a list of dictionaries of short and long reasons,
                    with keys 'short' and 'long'
    :type reasons: list of dicts
    :raises TypeError: if the parameter is not a list
    :raises KeyError: if the dictionaries in the list do not have the correct structure
    """

    conn, curs = get_reasons()

    if not isinstance(reasons, list):
        raise TypeError('reasons is not a list')

    try:
        for reason in reasons:
            if reason['short'] == DEFAULT_SHORT:
                continue
            curs.execute('SELECT shortreason FROM reasons WHERE shortreason=?', (reason['short'],))
            if not curs.fetchone():
                curs.execute('INSERT INTO reasons VALUES (?,?)', (reason['short'], reason['long'],))
        conn.commit()
    except KeyError:
        cherrypy.log('Parameter does not have correct keys.')
        raise

    conn.close()


def short_reasons_list():
    """Get the list of short reasons

    :returns: the list of short reasons
    :rtype: list of strs
    """

    conn, curs = get_reasons()

    curs.execute('SELECT shortreason FROM reasons')
    short_list = []
    for item, in curs.fetchall():
        short_list.append(item)

    conn.close()

    return short_list


def reasons_list():
    """Get the full list of reasons

    :returns: all of the reasons in a dictionary with the short reasons being the key
    :rtype: dict
    """

    short_list = short_reasons_list()

    conn, curs = get_reasons()

    reasons = {}

    for short in short_list:
        curs.execute('SELECT longreason FROM reasons WHERE shortreason=?', (short,))
        longreason = ''
        for item in curs.fetchall():
            longreason += (item[0])

        reasons[short] = longreason

    conn.close()

    return reasons
